from src.retrievals.losses.arcface import ArcFaceAdaptiveMarginLoss
from src.retrievals.losses.cosine_similarity import CosineSimilarity
from src.retrievals.losses.infonce import InfoNCE
from src.retrievals.losses.simcse import SimCSE
from src.retrievals.losses.triplet import TripletLoss
